# type: ignore

from access_py_telemetry.decorators import ipy_register_func, register_func
from access_py_telemetry.registry import TelemetryRegister
from access_py_telemetry.api import ApiHandler
import pytest
import asyncio


def test_ipy_register_func(api_handler, reset_telemetry_register):
    """
    Make sure that the decorator registers the function correctly.
    """

    @ipy_register_func(
        service="intake_catalog",
        extra_fields={"model": "ACCESS-OM2", "random_number": 2},
        pop_fields=["session_id"],
    )
    def my_func():
        pass

    my_func()

    register = TelemetryRegister("intake_catalog")
    api_handler = ApiHandler()
    blank_registries = {
        key: {} for key in api_handler.endpoints if key != "intake_catalog"
    }

    assert api_handler.extra_fields == {
        "intake_catalog": {"model": "ACCESS-OM2", "random_number": 2},
        **blank_registries,
    }

    assert api_handler.pop_fields == {"intake_catalog": ["session_id"]}

    assert my_func.__name__ in register

    # Reset the register to avoid breaking other tests

    register.deregister(my_func.__name__)


@pytest.mark.asyncio
async def test_register_func(api_handler, reset_telemetry_register):
    """
    Use the register_func decorator factory to register a function.
    """

    @register_func(
        service="intake_catalog",
        extra_fields={"model": "ACCESS-OM2", "random_number": 2},
        pop_fields=["session_id"],
    )
    def my_func():
        pass

    register = TelemetryRegister("intake_catalog")
    api_handler = ApiHandler()

    blank_registries = {
        key: {} for key in api_handler.endpoints if key != "intake_catalog"
    }

    assert api_handler.extra_fields == {
        "intake_catalog": {"model": "ACCESS-OM2", "random_number": 2},
        **blank_registries,
    }

    assert api_handler.pop_fields == {"intake_catalog": ["session_id"]}

    assert my_func.__name__ in register

    my_func()
    await asyncio.sleep(0)

    # Not quiet how to test this part of the code right now...

    # api_handler.server_url = "http://test_server.com"
    # with pytest.warns(RuntimeWarning):
    #     my_func()
    #     await asyncio.sleep(0)

    # Shtudown the event loop

    # Reset the register to avoid breaking other tests

    register.deregister(my_func.__name__)



#!/usr/bin/env python
# type: ignore

"""Tests for `access_py_telemetry` package."""

from access_py_telemetry.api import (
    SessionID,
    ApiHandler,
    ProductionToggle,
    send_in_loop,
    send_telemetry,
    _format_endpoint,
    _run_event_loop,
)
from unittest import mock
from pydantic import ValidationError
import pytest
from pytest_httpserver import HTTPServer, RequestMatcher
import time


@pytest.fixture
def local_host():
    return "http://localhost:8000"


@pytest.fixture
def default_url():
    return "https://reporting.access-nri-store.cloud.edu.au"


def test_session_id_properties():
    """
    Check that the SessionID class is a lazily evaluated singleton.
    """
    id1 = SessionID()

    assert hasattr(SessionID, "_instance")

    id2 = SessionID()

    assert id1 is id2

    assert type(id1) is str

    assert len(id1) == 36

    assert id1 != SessionID.create_session_id()


def test_api_handler_server_url(local_host, default_url, api_handler):
    """
    Check that the APIHandler class is a singleton.
    """

    session1 = api_handler
    session2 = ApiHandler()

    assert session1 is session2

    # Check defaults haven't changed by accident
    assert session1.server_url == default_url

    # Change the server url
    session1.server_url = local_host
    assert session2.server_url == local_host

    # ApiHandler._instance = None


def test_api_handler_extra_fields(local_host, api_handler):
    """
    Check that adding extra fields to the APIHandler class works as expected.
    """

    session1 = api_handler
    session2 = ApiHandler()

    session1.server_url = local_host
    assert session2.server_url == local_host

    # Change the extra fields - first
    with pytest.raises(AttributeError):
        session1.extra_fields = {"catalog_version": "1.0"}

    XF_NAME = "intake_catalog"

    session1.add_extra_fields(XF_NAME, {"version": "1.0"})

    blank_registries = {key: {} for key in session1.endpoints if key != XF_NAME}

    assert session2.extra_fields == {
        "intake_catalog": {"version": "1.0"},
        **blank_registries,
    }

    with pytest.raises(KeyError) as excinfo:
        session1.add_extra_fields("catalog", {"version": "2.0"})
        assert str(excinfo.value) == "Endpoint catalog not found"

    # Make sure that adding a new sesson doesn't overwrite the old one
    session3 = ApiHandler()
    assert session3 is session1
    assert session1.server_url == local_host
    assert session3.server_url == local_host


def test_api_handler_extra_fields_validation(api_handler):
    """
    Pydantic should make sure that if we try to update the extra fields, we have
    to pass the correct types, and only let us update fields through the
    add_extra_field method.
    """

    # Mock a couple of extra services

    api_handler.endpoints = {
        "catalog": "intake/update",
        "payu": "payu/update",
    }

    with pytest.raises(AttributeError):
        api_handler.extra_fields = {
            "catalog": {"version": "1.0"},
            "payu": {"version": "1.0"},
        }

    with pytest.raises(KeyError):
        api_handler.add_extra_fields("catalogue", {"version": "2.0"})

    with pytest.raises(ValidationError):
        api_handler.add_extra_fields("catalog", ["invalid", "type"])

    api_handler.add_extra_fields("payu", {"model": "ACCESS-OM2", "random_number": 2})


def test_api_handler_remove_fields(api_handler):
    """
    Check that we can remove fields from the telemetry record.
    """

    # Pretend we only have catalog & payu services and then mock the initialisation
    # of the _extra_fields attribute

    api_handler.endpoints = {
        "catalog": "intake/update",
        "payu": "payu/update",
    }

    api_handler._extra_fields = {
        ep_name: {} for ep_name in api_handler.endpoints.keys()
    }

    # Payu wont need a 'session_id' field, so we'll remove it

    api_handler.remove_fields("payu", ["session_id"])

    api_handler.add_extra_fields("payu", {"model": "ACCESS-OM2", "random_number": 2})

    payu_record = api_handler._create_telemetry_record(
        service_name="payu", function_name="_test", args=[], kwargs={}
    )
    payu_record["name"] = "test_username"

    assert payu_record == {
        "function": "_test",
        "args": [],
        "kwargs": {},
        "name": "test_username",
        "model": "ACCESS-OM2",
        "random_number": 2,
    }

    assert api_handler._pop_fields == {"payu": ["session_id"]}

    # Now remove the 'model' field from the payu record, as a string.
    api_handler.remove_fields("payu", "model")


@mock.patch("access_py_telemetry.api.getpass.getuser")
def test_api_handler_send_api_request(mock_getuser, api_handler, capfd):
    """
    Create and send an API request with telemetry data - just to make sure that
    the request is being sent correctly.
    """

    # Set the private production toggle to false, so that we don't invoke the hook
    # that changes the server url to the staging server
    api_handler.server_url = "http://dud/host/endpoint"

    # Set the username to `test_username` for testing purposes
    mock_getuser.return_value = "test_username"
    # Pretend we only have catalog & payu services and then mock the initialisation
    # of the _extra_fields attribute

    api_handler.endpoints = {
        "catalog": "intake/update",
        "payu": "payu/update",
    }

    api_handler._extra_fields = {
        ep_name: {} for ep_name in api_handler.endpoints.keys()
    }

    api_handler.add_extra_fields("payu", {"model": "ACCESS-OM2", "random_number": 2})

    # Remove indeterminate fields
    api_handler.remove_fields("payu", ["session_id"])

    api_handler.send_api_request(
        service_name="payu",
        function_name="_test",
        args=[1, 2, 3],
        kwargs={"random": "item"},
    )
    assert api_handler._last_record == {
        "function": "_test",
        "args": [1, 2, 3],
        "kwargs": {"random": "item"},
        "model": "ACCESS-OM2",
        "random_number": 2,
        "name": "test_username",
    }


def test_api_handler_invalid_endpoint(api_handler):
    """
    Create and send an API request with telemetry data.
    """

    # Pretend we only have catalog & payu services and then mock the initialisation
    # of the _extra_fields attribute

    with pytest.raises(KeyError) as excinfo:
        api_handler._get_endpoints(
            service_name="payu",
        )

    assert "Endpoint for 'payu' not found " in str(excinfo.value)


def test_send_in_loop_is_bg(httpserver: HTTPServer):
    """
    Send a request, but make sure that it runs in the background (ie. is non-blocking).

    There will be some overhead associated with the processes startup and teardown,
    but we shouldn't be waiting for the requests to finish. Using a long timeout
    and only sending 3 requests should be enough to ensure that we're not accidentally
    testing the process startup/teardown time.
    """
    httpserver.expect_request("/loop-endpoint").respond_with_data("Request received")
    httpserver.expect_request("/slow-endpoint").respond_with_handler(time.sleep(2))

    assert len(httpserver.log) == 0

    for _ in range(3):
        send_in_loop(
            endpoint=httpserver.url_for("/loop-endpoint"), telemetry_data={}, timeout=3
        )

    httpserver.assert_request_made(RequestMatcher("/loop-endpoint"), count=3)
    assert len(httpserver.log) == 3

    start_time = time.time()
    for _ in range(3):
        send_in_loop(
            endpoint=httpserver.url_for("/sleep-endpoint"), telemetry_data={}, timeout=3
        )

    end_time = time.time()

    dt = end_time - start_time

    assert dt < 4
    assert len(httpserver.log) == 6


def test_api_handler_set_timeout(api_handler):
    """
    Make sure that we can set the timeout for the APIHandler class, and that it
    is either a positive float or None.
    """

    with pytest.raises(ValueError):
        api_handler.request_timeout = -1

    with pytest.raises(TypeError):
        api_handler.request_timeout = "string"

    api_handler.request_timeout = 1.0

    assert api_handler.request_timeout == 1.0

    api_handler.request_timeout = None

    assert api_handler.request_timeout is None


def test_api_handler_url_warnings(api_handler):
    """
    Make sure that we get a warning if we try to set the server_url to a non-reporting
    url.
    """

    with pytest.warns(
        UserWarning, match="Server URL not an ACCESS-NRI Reporting API URL"
    ):
        api_handler.server_url = "http://localhost:8000"

    with pytest.warns(
        UserWarning, match="Server URL does not end with 'api' or 'api/'"
    ):
        api_handler.server_url = "https://localhost:8000"


def test_api_handler_url_no_warnings(api_handler, recwarn):
    """
    If we set NRI_USER to False, we shouldn't get any warnings - other orgs/users
    can do what they like.
    """
    assert len(recwarn) == 0

    import access_py_telemetry.api

    access_py_telemetry.api.NRI_USER = False
    # Would trigger two warnings if NRI_USER was True
    api_handler.server_url = "http://localhost:8000"
    # Clean this up
    access_py_telemetry.api.NRI_USER = True

    assert len(recwarn) == 0


@pytest.mark.parametrize(
    "server_url, endpoint, expected",
    [
        (
            "http://localhost:8000",
            "/some/endpoint",
            "http://localhost:8000/some/endpoint/",
        ),
        (
            "http://localhost:8000/",
            "some/endpoint/",
            "http://localhost:8000/some/endpoint/",
        ),
        (
            "https://localhost:8000",
            "/some/endpoint",
            "https://localhost:8000/some/endpoint/",
        ),
        (
            "https://localhost:8000/",
            "some/endpoint/",
            "https://localhost:8000/some/endpoint/",
        ),
    ],
)
def test_format_endpoint(server_url, endpoint, expected):
    assert _format_endpoint(server_url, endpoint) == expected


def test_api_handler_add_generic_token(api_handler):
    """
    Add a token to the APIHandler class, without specifying a service, and make
    sure it's applied to each service.
    """
    api_handler.set_headers(None, {"generic_token": "password123"})

    assert api_handler.headers == {
        endpoint: {"generic_token": "password123"} for endpoint in api_handler.endpoints
    }

    api_handler.clear_headers(None)

    assert api_handler.headers == {endpoint: {} for endpoint in api_handler.endpoints}


def test_api_handler_add_single_service_token(api_handler):
    """
    Add a token to the APIHandler class, specifying a service, and make sure it's
    only applied to that service.
    """
    endpoints = [k for k in api_handler.endpoints.keys()]
    specified_service, *unspecified_services = endpoints
    api_handler.set_headers(specified_service, {"catalog_token": "password123"})

    assert api_handler.headers == {
        specified_service: {"catalog_token": "password123"},
        **{service: {} for service in unspecified_services},
    }

    api_handler.clear_headers("intake_catalog")

    assert api_handler.headers == {endpoint: {} for endpoint in api_handler.endpoints}


def test_api_handler_sends_correct_headers(api_handler, httpserver):
    api_handler.server_url = httpserver.url_for("/")
    api_handler.endpoints = {"endpoint": "endpoint"}
    api_handler.set_headers(None, {"generic_token": "password123"})
    httpserver.expect_request(
        "/endpoint", headers={"generic_token": "password123"}
    ).respond_with_data("Request received")
    api_handler.send_api_request(
        service_name="endpoint",
        function_name="test",
        args=[1, 2, 3],
        kwargs={"random": "item"},
    )


def test_api_handler_raises_invalid_service_headers(api_handler):
    """
    Make sure that we can't set headers for a service that doesn't exist.
    """
    with pytest.raises(KeyError):
        api_handler.set_headers("invalid_service", {"catalog_token": "password123"})

    with pytest.raises(KeyError):
        api_handler.clear_headers("invalid_service")


def test__run_event_loop(httpserver):
    """
    USe the _run_event_loop function to send a request to a server, and make sure
    that the request is sent correctly.
    """
    _run_event_loop(httpserver.url_for("/"), {}, {})
    httpserver.assert_request_made(RequestMatcher("/"), count=1)


def test_production_toggle(production_toggle, api_handler):
    """
    Make sure that the production toggle works as expected.
    """

    # Check that the production toggle is set to True by default
    assert production_toggle._production
    assert production_toggle.production

    s = str(production_toggle)
    assert s == "ProductionToggle(production=True)"

    # Check that the production toggle can be set to False
    production_toggle.production = False

    assert not production_toggle._production
    assert not production_toggle.production
    assert api_handler.server_url == ProductionToggle.STAGING_URL

    # Check that the production toggle can be set back to True
    production_toggle.production = True

    assert production_toggle._production
    assert production_toggle.production
    assert api_handler.server_url == ProductionToggle.PRODUCTION_URL

    with pytest.raises(TypeError):
        production_toggle.production = 1


@pytest.mark.parametrize(
    "warn_set, response, raise_warning",
    [
        (  # No warning on 200
            True,
            {"response_data": "Success!", "status": 200},
            False,
        ),
        (  # Warn on 403, if warn_set is True
            True,
            {"response_data": "Auth Failure!", "status": 403},
            True,
        ),
        (  # Dont warn on 403, if warn_set is False
            False,
            {"response_data": "Auth Failure!", "status": 403},
            False,
        ),
        (  # Warn on 404, if warn_set is True
            True,
            {"response_data": "Auth Failure!", "status": 404},
            True,
        ),
        (  # Dont warn on 404, if warn_set is False
            False,
            {"response_data": "Auth Failure!", "status": 404},
            False,
        ),
        (  # Warn on 500, if warn_set is True
            True,
            {"response_data": "Auth Failure!", "status": 500},
            True,
        ),
        (  # Dont warn on 500, if warn_set is False
            False,
            {"response_data": "Auth Failure!", "status": 500},
            False,
        ),
    ],
)
@pytest.mark.asyncio
async def test_send_telemetry_warning_toggle(
    httpserver: HTTPServer, warn_set, response, raise_warning
):
    """
    Send a request, but make sure that it runs in the background (ie. is non-blocking).

    There will be some overhead associated with the processes startup and teardown,
    but we shouldn't be waiting for the requests to finish. Using a long timeout
    and only sending 3 requests should be enough to ensure that we're not accidentally
    testing the process startup/teardown time.
    """
    httpserver.expect_request("/endpoint").respond_with_data(**response)

    if raise_warning:
        with pytest.warns(RuntimeWarning, match="Request failed"):
            await send_telemetry(
                endpoint=httpserver.url_for("/endpoint"),
                data={},
                headers={},
                warn=warn_set,
            )
    else:
        await send_telemetry(
            endpoint=httpserver.url_for("/endpoint"),
            data={},
            headers={},
            warn=warn_set,
        )



from typing import Callable, Any, Iterable
from .registry import TelemetryRegister
from functools import wraps

from .api import ApiHandler, send_in_loop


def ipy_register_func(
    service: str,
    extra_fields: dict[str, Any] | None = None,
    pop_fields: Iterable[str] | None = None,
) -> Callable[..., Any]:
    """
    Decorator to register a function in the specified service and track usage
    using IPython events. This hides a lot of complexity which is more visible in
    the `register_func` decorator.

    Parameters
    ----------
    func : Callable
        The function to register.
    service : str
        The name of the telemetry register to use.
    extra_fields : Iterable[dict[str, Any]], optional
        Extra fields to add to the telemetry record. These can also be added after
        the fact using the `add_extra_field` method.
    pop_fields : Iterable[str], optional
        Fields to remove from the telemetry record. This can be useful for removing
        default fields that are not needed for a particular function

    Returns
    -------
    Callable
        The function with the telemetry decorator.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        api_handler = ApiHandler()

        api_handler.add_extra_fields(service, extra_fields or {})
        api_handler.remove_fields(service, pop_fields or [])
        TelemetryRegister(service).register(func.__name__)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return wrapper

    return decorator


def register_func(
    service: str,
    extra_fields: dict[str, Any] | None = None,
    pop_fields: Iterable[str] | None = None,
) -> Callable[..., Any]:
    """
    Decorator to register a function in the specified service and track usage
    with async requests.

    Parameters
    ----------
    func : Callable
        The function to register.
    service : str
        The name of the telemetry register to use.
    extra fields : Iterable[dict[str, Any]], optional
        Extra fields to add to the telemetry record. These can also be added after
        the fact using the `add_extra_field` method.
    pop_fields : Iterable[str], optional
        Fields to remove from the telemetry record. This can be useful for removing
        default fields that are not needed for a particular function.

    Returns
    -------
    Callable
        The function with the telemetry decorator.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        api_handler = ApiHandler()

        # Configure fields & register the function
        api_handler.add_extra_fields(service, extra_fields or {})
        api_handler.remove_fields(service, pop_fields or [])
        TelemetryRegister(service).register(func.__name__)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            telemetry_data = api_handler._create_telemetry_record(
                service, func.__name__, args, kwargs
            )

            endpoint = f"{api_handler.server_url}{api_handler.endpoints[service]}"

            print(f"Sending telemetry data to {endpoint}")

            send_in_loop(endpoint, telemetry_data)

            # Call the original function
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
Copyright 2022 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
SPDX-License-Identifier: Apache-2.0
"""

from typing import Any
import ast
import re
from IPython.core.getipython import get_ipython
from IPython.core.interactiveshell import ExecutionInfo

from .api import ApiHandler
from .registry import TelemetryRegister
from .utils import REGISTRIES

api_handler = ApiHandler()

registries = {registry: TelemetryRegister(registry) for registry in REGISTRIES.keys()}


def strip_magic(code: str) -> str:
    """
    Parse the provided code into an AST (Abstract Syntax Tree).

    Parameters
    ----------

    code : str
        The code to parse.
    Returns
    -------
    str
        The code without IPython magic commands.

    """

    IPYTHON_MAGIC_PATTERN = r"^\s*[%!?]{1,2}|^.*\?{1,2}$"

    code = "\n".join(
        line for line in code.splitlines() if not re.match(IPYTHON_MAGIC_PATTERN, line)
    )

    return code


def capture_registered_calls(info: ExecutionInfo) -> None:
    """
    Use the AST module to parse the code that we are executing & send an API call
    if we detect specific function or method calls.

    Fail silently if we can't parse the code.

    Parameters
    ----------
    info : IPython.core.interactiveshell.ExecutionInfo
        An object containing information about the code being executed.

    Returns
    -------
    None
    """
    code = info.raw_cell

    if code is None:
        return None

    code = strip_magic(code)

    try:
        tree = ast.parse(code)
    except (SyntaxError, IndentationError):
        return None

    user_namespace: dict[str, Any] = get_ipython().user_ns  # type: ignore

    try:
        visitor = CallListener(user_namespace, REGISTRIES, api_handler)
        visitor.visit(tree)
    except Exception:
        # Catch all exceptions to avoid breaking the execution
        # of the code being run.
        return None

    return None


class CallListener(ast.NodeVisitor):
    def __init__(
        self,
        user_namespace: dict[str, Any],
        registries: dict[str, set[str]],
        api_handler: ApiHandler,
    ):
        self.user_namespace = user_namespace
        self.registries = registries
        self._caught_calls: set[str] = set()  # Mostly for debugging
        self.api_handler = api_handler

    def _get_full_name(self, node: ast.AST) -> str | None:
        """Recursively get the full name of a function or method call."""
        if isinstance(node, ast.Attribute):
            return f"{self._get_full_name(node.value)}.{node.attr}"
        elif isinstance(node, ast.Name):
            return node.id
        return None

    def safe_eval(self, node: ast.AST) -> Any:
        """Try to evaluate a node, or return the unparsed node if that fails."""
        try:
            return ast.literal_eval(node)
        except (ValueError, SyntaxError):
            return ast.unparse(node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        if full_name := self._get_full_name(node):
            self._process_api_call(full_name, [], {})

        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        full_name = self._get_full_name(node.func)
        func_name = None
        if full_name:
            parts = full_name.split(".")
            if len(parts) == 1:
                # Regular function call
                func_name = f"{full_name}"
            else:
                # Check if the first part is in the user namespace
                instance = self.user_namespace.get(parts[0])
                if instance is None:
                    self.generic_visit(node)
                    return None

                class_name = type(instance).__name__
                if class_name != "module":
                    func_name = f"{class_name}.{'.'.join(parts[1:])}"
                else:
                    func_name = f"{instance.__name__}.{'.'.join(parts[1:])}"

        args = [self.safe_eval(arg) for arg in node.args]
        kwargs = {
            kw.arg: self.safe_eval(kw.value)
            for kw in node.keywords
            if kw.arg is not None
        }

        if func_name:
            self._process_api_call(func_name, args, kwargs)

        self.generic_visit(node)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        """Handle subscript operations."""
        full_name = self._get_full_name(node.value)  # Get the object being indexed
        func_name = None
        if full_name:
            parts = full_name.split(".")
            instance = self.user_namespace.get(parts[0])
            if instance is None:
                return None

            class_name = type(instance).__name__
            func_name = f"{class_name}.{'.'.join(parts[1:])}__getitem__"

        if isinstance(node.slice, ast.Name):
            args = self.user_namespace.get(node.slice.id, None)
        else:
            args = ast.unparse(node.slice)

        if func_name:
            self._process_api_call(func_name, [args], {})

        self.generic_visit(node)

    def _process_api_call(
        self, func_name: str, args: list[Any], kwargs: dict[str, Any]
    ) -> None:
        """Process an API call for a matched function name."""
        for registry, registered_funcs in self.registries.items():
            if func_name in registered_funcs:
                self.api_handler.send_api_request(
                    registry,
                    func_name,
                    args,
                    kwargs,
                )
                self._caught_calls |= {func_name}



# type: ignore
import pytest
from unittest import mock

from access_py_telemetry.cli import configure_telemetry


@pytest.fixture
def test_dir(tmp_path):
    return tmp_path


@pytest.mark.parametrize(
    "argv, expected_output, expected_ret_code",
    [
        (["--disable"], "Telemetry already disabled.\n", 0),
        (["--enable"], "Telemetry enabled.\n", 0),
        (["--status"], "Telemetry disabled.\n", 0),
        (
            ["--disable", "--enable"],
            "Only one of --disable, --enable, or --status can be used at a time.\n",
            1,
        ),
    ],
)
@mock.patch("pathlib.Path.home")
def test_configure_telemetry(
    home, test_dir, capsys, argv, expected_output, expected_ret_code
):
    home.return_value = test_dir

    ret_code = configure_telemetry(argv)
    output = capsys.readouterr()

    assert output.out == expected_output
    assert ret_code == expected_ret_code


@mock.patch("pathlib.Path.home")
def test_telemetry_configure_helpmsg(home, test_dir, capsys):
    home.return_value = test_dir

    ret_code = configure_telemetry([])
    output = capsys.readouterr()

    assert (
        "[-h] [--disable] [--enable] [--status] [--silent]\n\nConfigure ipython telemetry.\n\noptions:\n  -h, --help  show this help message and exit\n  --disable   Disable telemetry.\n  --enable    Enable telemetry.\n  --status    Check telemetry status.\n  --silent    Suppress output.\n"
        in output.out
    )

    assert ret_code == 0


@pytest.mark.parametrize(
    "argv1, argv2, expected_output_1, expected_output_2,",
    [
        (
            ["--enable"],
            ["--status"],
            "Telemetry enabled.\n",
            "Telemetry enabled.\n",
        ),
        (
            ["--enable"],
            ["--disable"],
            "Telemetry enabled.\n",
            "Telemetry disabled.\n",
        ),
        (
            ["--disable"],
            ["--status"],
            "Telemetry already disabled.\n",
            "Telemetry disabled.\n",
        ),
        (
            ["--enable"],
            ["--enable"],
            "Telemetry enabled.\n",
            "Telemetry already enabled.\n",
        ),
    ],
)
@mock.patch("pathlib.Path.home")
def test_telemetry_runtwice(
    home, test_dir, capsys, argv1, argv2, expected_output_1, expected_output_2
):
    home.return_value = test_dir

    configure_telemetry(argv1)
    output_1 = capsys.readouterr()

    configure_telemetry(argv2)
    output_2 = capsys.readouterr()

    assert output_1.out == expected_output_1
    assert output_2.out == expected_output_2


@mock.patch("pathlib.Path.home")
def test_telemetry_corrupted(home, test_dir, capsys):
    """
    After enabling telemetry, we will corrupt the telemetry file and check if the status is reported correctly.
    """
    home.return_value = test_dir

    configure_telemetry(["--enable"])
    output_1 = capsys.readouterr()

    telemetry_fname = (
        test_dir / ".ipython" / "profile_default" / "startup" / "telemetry.py"
    )

    with open(telemetry_fname, "w") as f:
        f.write("Corrupted file")

    configure_telemetry(["--status"])
    output_2 = capsys.readouterr()

    assert output_1.out == "Telemetry enabled.\n"
    assert (
        output_2.out
        == "Telemetry enabled but misconfigured. Run `access-py-telemetry --disable && access-py-telemetry --enable` to fix.\n"
    )



"""
Copyright 2022 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
SPDX-License-Identifier: Apache-2.0
"""

from typing import Any
import yaml
from pathlib import Path
from dataclasses import dataclass, field

with open(Path(__file__).parent / "config.yaml", "r") as f:
    config = yaml.safe_load(f)


@dataclass
class TelemetryRegister:
    endpoint: str
    items: set[str] = field(default_factory=set)


def build_endpoints(
    config: dict[str, Any], parent: str | None = None
) -> list[TelemetryRegister]:
    """
    Recursively join the keys of the dictionary until we reach a list
    Returns list of tuples (path, endpoint_list)
    """
    parent = parent or ""
    results = []

    for key, val in config.items():
        if isinstance(val, dict):
            # Recursively process dictionaries and extend results
            nested_results = build_endpoints(val, f"{parent}/{key}" if parent else key)
            results.extend(nested_results)
        elif isinstance(val, list):
            # Add tuple of (path, list) when we find a list
            full_path = "/".join([parent, key]) if parent else key
            results.append(TelemetryRegister(full_path, set(val)))

    return results


ENDPOINTS = {
    register.endpoint.replace("/", "_"): register.endpoint
    for register in build_endpoints(config)
}

REGISTRIES = {
    register.endpoint.replace("/", "_"): register.items
    for register in build_endpoints(config)
}



"""
Copyright 2022 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
SPDX-License-Identifier: Apache-2.0
"""

from typing import Type, TypeVar, Iterator, Callable, Any
import pydantic
import copy
from .utils import REGISTRIES

T = TypeVar("T", bound="TelemetryRegister")


class RegisterWarning(UserWarning):
    """
    Custom warning class for the TelemetryRegister class.
    """

    pass


class TelemetryRegister:
    """
    Singleton class to register functions for telemetry. Like the session handler,
    this class is going to be a singleton so that we can register functions to it
    from anywhere and have them persist across all telemetry calls.

    """

    # Set of registered functions for now - we can add more later or dynamically
    # using the register method.

    _instances: dict[str, "TelemetryRegister"] = {}

    def __new__(cls: Type[T], service: str) -> T:
        if cls._instances.get(service) is None:
            cls._instances[service] = super().__new__(cls)
        return cls._instances[service]  # type: ignore

    def __init__(self, service: str) -> None:
        if hasattr(self, "_initialized"):
            return None
        self._initialized = True
        self.service = service
        self.registry = copy.deepcopy(REGISTRIES.get(service, set()))

    def __str__(self) -> str:
        return str(list(self.registry))

    def __repr__(self) -> str:
        """
        I'm going to cheat and just print out the registry for now.
        """
        return self.__str__()

    def __contains__(self, function_name: str) -> bool:
        return function_name in self.registry

    def __iter__(self) -> Iterator[str]:
        return iter(self.registry)

    @pydantic.validate_call
    def register(self, *func_names: str | Callable[..., Any]) -> None:
        """
        Register functions to the telemetry register.


        Parameters
        ----------
        func_names : Sequence[str | Callable]
            The name of the function to register. Can also be a list of function names.
            If you pass a function, it will register the function by name, using
            the __name__ attribute.


        Returns
        -------
        None
        """

        for func in func_names:
            if isinstance(func, str):
                self.registry.add(func)
            else:
                self.registry.add(func.__name__)
        return None

    @pydantic.validate_call
    def deregister(self, *func_names: str | Callable[..., Any]) -> None:
        """
        Deregister a function from the telemetry register.

        Parameters
        ----------
        function_name : str
            The name of the function to deregister. Can also be a list of function names.
            If you pass a function, it will deregister the function by name, using
            the __name__ attribute.

        Returns
        -------
        None
        """
        for func in func_names:
            if isinstance(func, str):
                self.registry.remove(func)
            else:
                self.registry.remove(func.__name__)
        return None
